function [H, msk, dm, dn]=rtncll(lvls, prd, ang, cst, K)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% [H, msk, dm, dn]=irtncll(lvls, prd, ang, cst, K)
% IrRaTioNal CeLL dither array (function)
%   Generates an rational cell dither array nearest the user
%       requested screen angle and period (inverse frequency) across a
%       single cell.
%
% INPUT ARGUMENTS:
%       lvls  -> The number of quantization levels. If lvls is set to 256,
%                the returned dither array is type uint8.
%       prd   -> The user requested screen period in pixels.
%       ang   -> The user requested screen angle in degrees.
%       cst   -> A cost function, to be called by means of feval, that
%                takes two input variables (x,y) and returns two cost
%                values (default is cart2pol()).
%       K     -> Number of halftone cells the dither should include.
%
% OUTPUT ARGUMENT:
%       H    -> The resulting dither array.
%       msk  -> Complex mask image that stores the horizontal (real) and
%               vertical (imag) halftone cell index.
%       dm   -> The vertical spacing in pixels across K neighboring 
%               halftone cell centroids.
%       dn   -> The horizontal spacing in pixels across K neighboring
%               halftone cell centroids.
%
% EXAMPLE:
%	H = rtncll(256, 10, 15, 'rndsptfn', 5);
%   imshow(H);
%
% Dr. Daniel L. Lau
% dllau@engr.uky.edu
%
% August 21, 2007
% Copyright 2007 Dr. Daniel L. Lau
%
% Last modified on November 26, 2007
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (nargin<5)
    K=1;
end;

if (nargin<4)
    cst=[];
end;

if (isempty(cst))
    cst='cart2pol';
end;

if (length(K)==1)
    dn=K*round(prd*cos(ang*pi/180.0));
    dm=K*round(prd*sin(ang*pi/180.0));
    prd=sqrt(dm*dm+dn*dn)/K;
    ang=atan2(dm, dn)/pi*180.0;
else
    dn=round(prd*cos(ang*pi/180.0));
    dm=round(prd*sin(ang*pi/180.0));    
    prd=sqrt(dm*dm+dn*dn);
    ang=atan2(dm, dn)/pi*180.0;
end;
    
if (dm>0 & dn>0)
    dmp=dm;
    dnp=dn;
elseif (dm>0 & dn<0)
    dmp=-dn;
    dnp=dm;
    ang=ang-90;
elseif (dm<0 & dn<0)
    dmp=-dm;
    dnp=-dn;
    ang=ang+180;
elseif (dm<0 & dn>0)
    dmp=dn;
    dnp=-dm;
    ang=ang+90;
end;

[N,M]=meshgrid([-dmp+1:dnp-1],[0:dmp+dnp-1]);

[a,r]=cart2pol(N,M);
[x,y]=pol2cart(a-ang*pi/180.0,r/prd);

x=round(x*1e8)/1e8;
y=round(y*1e8)/1e8;

xpp=floor(x);
xp=2.0*(x-xpp)-1.0;
ypp=floor(y);
yp=2.0*(y-ypp)-1.0;

H=zeros(size(xp));
msk=zeros(size(H));
l=find((ypp==0)&(xpp==0));

xa=xp(l);
ya=yp(l);

[r,a]=feval(cst, xa, ya);
[a,i]=sort(a);
[r,j]=sort(r(i));
i=i(j);

levels=floor([0:length(l)-1]'/(length(l)-1)*(lvls-1));

H(l(i))=levels;

j=1;
for m=0:K-1
    for n=0:K-1
        s=find((ypp==m)&(xpp==n));
        msk(s)=j;
        j=j+1;
        
        if (m==0 & n==0) continue; end;
        H(s)=H(l);
    end;
end;

xp=xp.*(msk~=0);
yp=yp.*(msk~=0);

if (dm>0 & dn<0)
    H=rot90(H,-1);
    msk=rot90(msk,-1);
    x=rot90(xp,-1);
    y=rot90(yp,-1);
elseif (dm<0 & dn<0)
    H=rot90(H,-2);
    msk=rot90(msk,-2);
    x=rot90(xp,-2);
    y=rot90(yp,-2);
elseif (dm<0 & dn>0)
    H=rot90(H,-3);
    msk=rot90(msk,-3);
    x=rot90(xp,-3);
    y=rot90(yp,-3);
else
    x=xp;
    y=yp;
end;

if (lvls==256)
    H=uint8(H);
end;

return;
