function Hp=zrnglrry(H, dm, dn, szLmt)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Hp=zrnglrry(dm, dn, H, szLmt)
% ZeRo aNGLe aRRay (function)
%   Converts the supplied dither array, along with the horizontal and
%       vertical spacing between neighboring halftone cells in the array,
%       to zero angle, square dither array with no gaps.
%
% INPUT ARGUMENTS:
%       H     -> The original dither array.
%       dm    -> The vertical spacing in pixels across K neighboring 
%                halftone cell centroids.
%       dn    -> The horizontal spacing in pixels across K neighboring
%                halftone cell centroids.
%       szLmt -> Maximum row and column size of the resulting zero-angle
%                dither array, provided for convenience for low-memory
%                computers.
%
% OUTPUT ARGUMENT:
%       Hp    -> Zero-angle dither array conversion of H.
%
% EXAMPLE:
%	[H, dm, dn] = rtncll(256, 10, 15, 'rndsptfn', 5);
%   Hp=zrnglrry(H, dm, dn);
%   imshow(Hp);
%
% Dr. Daniel L. Lau
% dllau@engr.uky.edu
%
% August 21, 2007
% Copyright 2007 Dr. Daniel L. Lau
%
% Last modified on November 26, 2007
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (nargin<4)
    szLmt=1000;
elseif (isempty(szLmt))
    szLmt=1e10;
end;

ngl=atan(dm/dn);

lng=sqrt(dm*dm+dn*dn);                      % length of one side of halftone cell

K=gcd(dm,dn);                               % get greatest common divisor
prd=sqrt((lng*dn/K)^2+(lng*dm/K)^2);        % find smallest period of zero angle dither array
prd=round(prd);

if (prd>szLmt)
    fprintf('Warning!, Dither array size exceeds size limit.\n');
    prd=szLmt;
end;

%%% NOW WE NEED TO GENERATE THE FINAL ZERO-ANGLE DITHER ARRAY
[N,M]=meshgrid([0:prd-1], [0:prd-1]);       % get row and column coordinates of zero angle dither array

[a,r]=cart2pol(N,M);                        % map these coordinates into rotated space
[X,Y]=pol2cart(a-ngl, r/lng);

X=round(X*1e8)/1e8;
Y=round(Y*1e8)/1e8;

Xp=X-floor(X);                      % give me coordinate within cell
Yp=Y-floor(Y);

[a,r]=cart2pol(Xp,Yp);              % map cell coordinate back into non-rotated space
[Np, Mp]=pol2cart(a+ngl, r*lng);
Mp=round(Mp);                       % round off coordinate into nearest integer
Np=round(Np);
Mp=Mp-min(min(Mp))+1;               % map this value to coordinate inside halftone cell
Np=Np-min(min(Np))+1;

if (isa(H, 'uint8'))
    Hp=zeros(size(M), 'uint8');     % now create the zero angle dither array
else
    Hp=zeros(size(M));          	% now create the zero angle dither array
end;

for m=1:length(Hp(:))
    Hp(m)=H(Mp(m), Np(m));
end;

%%% NOW WE NEED TO GENERATE THE MASK IMAGE CONTAINING CELLS IDS FOR EACH PIXEL
Xp=floor(X);
Yp=floor(Y);

[t,r]=cart2pol(Xp,Yp);
r=round(r*1e8)/1e8;

[Mp,Np]=pol2cart(t+ngl, r*lng);
Mp=round(Mp);
l=find(Mp<0);
Mp(l)=Mp(l)+prd;
l=find(Mp>=prd);
Mp(l)=Mp(l)-prd;

Np=round(Np);
l=find(Np<0);
Np(l)=Np(l)+prd;
l=find(Np>=prd);
Np(l)=Np(l)-prd;

msk=Mp+i*Np;

%%% NOW LETS CALCULATE THE WIDTH OF THE DITHER ARRAY IN UNITS OF HALFTONE CELLS
K=lng*lng;

return;
