/************************************************************************************/
/*                                                                                  */
/* ERRor_DIFfusion_HYSterisis                                                       */
/*  This function performs the error diffusion algorithm using a hysterisis         */
/*  constant.  The original gray-scale image (0=black, 1=white) is passed           */
/*  along with the filter coefficients stored in a weight matrix such that the      */
/*  number of rows and columns in the weight matrix are odd, and the center         */
/*  index corresponding to the pixel being halftoned.  All filter weights after     */
/*  this pixel correspond to the diffusion of error, while all previous weights     */
/*  correspond to the hysterisis term.                                              */
/*                                                                                  */
/* Synopsis:                                                                        */
/*  Y=halftone(X, W)                                                                */
/*      Y=Output image (binary)                                                     */
/*      X=Input image (intensity)                                                   */
/*      W=Weighting matrix with number of rows=2*M+1 and                            */
/*        number of columns=2*N+1 such that                                         */
/*        center pixel is W(M,N)                                                    */
/*                                                                                  */
/*      Y=halftone(X, W, H)                                                         */
/*      H=Hysterisis constant, if H is a matrix of size Hm, Hn                      */
/*        then the output at X[m,n] is determined using                             */
/*        hysterisis constant H[m%Hm, n%Hn].                                        */
/*  Y=halftone(X, W, H, d)                                                          */
/*      d=raster scan(1, default) or serpentine scan(-1)                            */
/*                                                                                  */
/*  Y=halftone(X, W, H, d, P)                                                       */
/*      P=Pair combinations for perturbing weights.                                 */
/*                                                                                  */
/*  Y=halftone(X, W, H, d, P, T)                                                    */
/*      T=Tolerances for perturbing weights, for matrix T, errdif uses the          */
/*        tolerance of the major weight.                                            */
/*                                                                                  */
/*  Y=halftone(X, W, H, d, P, T, s)                                                 */
/*      s=Seed for random number generator.                                         */
/*                                                                                  */
/* Daniel Leo Lau                                                                   */
/* Copyright February 22, 1997                                                      */
/*                                                                                  */
/************************************************************************************/ 
#include <math.h>
#include "mex.h"
#define White_Value             1.0
#define Threshold               0.5
#define Hysterisis_Constant     1.0

int direction;
int image_row;
int image_col;
int weight_row;
int weight_col;
int hyst_row;
int hyst_col;
int num_filter_pairs;
double *weight_matrix;
double *hysterisis;

struct filter_pair{
    int key;
    int major;
    int minor;
    double tolerance;
} *filter_pairs;

void calculate_random_weights(double[]);
void error_diffusion_with_hysterisis(double*, double*, double*);
void initialize_filter_pairs(double*, double*);
int in_key_list(double);

void calculate_random_weights(double new_weights[])
{
    int m;
    double rn;

    for (m=0; m<num_filter_pairs; m++){
        rn=(((double)rand()/(double)RAND_MAX)-0.5)*2.0*filter_pairs[m].tolerance*weight_matrix[filter_pairs[m].minor];
        new_weights[filter_pairs[m].major]=weight_matrix[filter_pairs[m].major]+rn;
        new_weights[filter_pairs[m].minor]=weight_matrix[filter_pairs[m].minor]-rn;
        }
    return;
}

void error_diffusion_with_hysterisis(double *output_image, double *error_image, double *input_image)
{   int m_p, n_p, r, t, d;
    int m, n, weight_row_radius, weight_col_radius;
    double *rweight_matrix;

    weight_row_radius=weight_row/2;
    weight_col_radius=weight_col/2;
    rweight_matrix=(double*)mxCalloc(weight_row*weight_col, sizeof(double));

    for (m=0; m<weight_row*weight_col; m++) rweight_matrix[m]=weight_matrix[m];
    for (m=0; m<image_row; m++){
        d=(int)pow((float)direction, (float)m);
        if (d==1) n=0; else n=image_col-1;
        while(n>=0 && n<image_col){
            calculate_random_weights(rweight_matrix);
            output_image[m+n*image_row]=0.0;
            for (r=-1*weight_row_radius; r<=0; r++){
                for (t=(-1*weight_col_radius); t <= (weight_col_radius); t++){
                    if (r==0 && t>=0) continue;
                    m_p=m+r;   if (m_p < 0 || m_p >= image_row) continue;
                    n_p=n+d*t; if (n_p < 0 || n_p >= image_col) continue;
                    output_image[m+n*image_row] += rweight_matrix[r+weight_row_radius+(t+weight_col_radius)*weight_row] * output_image[m_p+n_p*image_row];
                    }
                }
            output_image[m+n*image_row]=output_image[m+n*image_row]*
                            hysterisis[(m%hyst_row)+(n%hyst_col)*hyst_row] +
                            input_image[m+n*image_row] + error_image[m+n*image_row];
            if (output_image[m+n*image_row] >= Threshold)
                output_image[m+n*image_row]=1.0;
            else
                output_image[m+n*image_row]=0.0;
            error_image[m+n*image_row]=error_image[m+n*image_row]+input_image[m+n*image_row]-output_image[m+n*image_row] * White_Value;
            for (r=0; r<=weight_row_radius; r++){
                for (t=-1*weight_col_radius; t<=weight_col_radius; t++){
                    if (r==0 && t<=0) continue;
                    m_p=m+r; if (m_p < 0 || m_p >= image_row) continue;
                    n_p=n+d*t; if (n_p < 0 || n_p >= image_col) continue;
                    error_image[m_p+n_p*image_row] += error_image[m+n*image_row] * rweight_matrix[r+weight_row_radius+(t+weight_col_radius)*weight_row];
                    }
                }
            n+=d;
            }
        }
    return;
}

int in_key_list(double key)
{
    int m;
    for (m=0; m<num_filter_pairs; m++){
        if (key==filter_pairs[m].key) return(1);
        }
    return(0);
}

void initialize_filter_pairs(double *pair_combinations, double *tolerances)
{
    int m,n,j;
    int invalid_flag;

    num_filter_pairs=0;
    for (m=0; m<weight_row*weight_col; m++){
        if (pair_combinations[m]==0) continue;
        invalid_flag=1;
        for (n=0; n<weight_row*weight_col; n++){
            if (m==n) continue;
            if (pair_combinations[m]==pair_combinations[n]){
                if (invalid_flag==0)
                    mexErrMsgTxt("Improper Pair Combination!");
                else
                    invalid_flag=0;
                }
            }
        if (invalid_flag==1) mexErrMsgTxt("Improper Pair Combination!");
        num_filter_pairs++;
        }
    num_filter_pairs=num_filter_pairs/2;
    filter_pairs=(struct filter_pair*)mxCalloc(num_filter_pairs, sizeof(struct filter_pair));
    j=0;
    for (m=0; m<weight_row*weight_col-1; m++){
        if (pair_combinations[m]==0) continue;
        if (!in_key_list(pair_combinations[m])){
            filter_pairs[j].key=pair_combinations[m];
            for (n=m+1; n<weight_row*weight_col; n++){
                if (pair_combinations[n]==pair_combinations[m]) break;
                }
            if (fabs(weight_matrix[m])<fabs(weight_matrix[n])){
                filter_pairs[j].major=n;
                filter_pairs[j].minor=m;
                filter_pairs[j].tolerance=tolerances[n];
                }
            else {
                filter_pairs[j].major=m;
                filter_pairs[j].minor=n;
                filter_pairs[j].tolerance=tolerances[m];
                }
            j++;
            }
        }
    return;
}

/*******************************************************************************/ 
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/ 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    unsigned char *output_data;
    double *output_image, *error_image, *input_image, *pair_combinations;
    double *tolerances;
    int m, n, output_dims[2];

    /****** Check for errors in user's call to function. ***********************/
    if (nrhs<2 || nrhs>7)
        mexErrMsgTxt("Error Diffusion requires two to six input arguments!");
    else if (nlhs>2)
        mexErrMsgTxt("Error Diffusion returns one or two output arguments!");
    else if (!mxIsNumeric(prhs[0]) ||
              mxIsComplex(prhs[0]) ||
              mxIsSparse(prhs[0])  ||
             !mxIsDouble(prhs[0]))
        mexErrMsgTxt("Input X must be a real matrix of type double!");
    else if (!mxIsNumeric(prhs[1]) ||
              mxIsComplex(prhs[1]) ||
              mxIsSparse(prhs[1])  ||
             !mxIsDouble(prhs[1]))
        mexErrMsgTxt("Input W must be a real matrix of type double!");
    else if (nrhs>=3 &&(!mxIsNumeric(prhs[2]) ||
                     mxIsComplex(prhs[2]) ||
                     mxIsSparse(prhs[2])  ||
                    !mxIsDouble(prhs[2])))
        mexErrMsgTxt("Input H must be a real matrix of type double!");
    else if (nrhs>=4 &&(!mxIsNumeric(prhs[3]) ||
                     mxIsComplex(prhs[3]) ||
                     mxIsSparse(prhs[3])  ||
                    !mxIsDouble(prhs[3])))
        mexErrMsgTxt("Input D must be a real scalar!");
    else if (nrhs>=5 && (!mxIsNumeric(prhs[4]) ||
              mxIsComplex(prhs[4]) ||
              mxIsSparse(prhs[4])  ||
             !mxIsDouble(prhs[4])))
        mexErrMsgTxt("Input P must be a real matrix of type double!");
    else if (nrhs>=5 && (mxGetM(prhs[4])!=mxGetM(prhs[1]) || mxGetN(prhs[4])!=mxGetN(prhs[1])))
        mexErrMsgTxt("Input P must be same size as W!");
    else if (nrhs>=6 && (!mxIsNumeric(prhs[5]) ||
                      mxIsComplex(prhs[5]) ||
                      mxIsSparse(prhs[5])  ||
                     !mxIsDouble(prhs[5])))
        mexErrMsgTxt("Input TOL must be a real matrix of type double!");
    else if (nrhs>=6 && (mxGetM(prhs[5])!=mxGetM(prhs[1]) || mxGetN(prhs[5])!=mxGetN(prhs[1])))
        mexErrMsgTxt("Input TOL must be same size as W!");
    else if (nrhs==7 && (!mxIsNumeric(prhs[6]) ||
                      mxIsComplex(prhs[6]) ||
                      mxIsSparse(prhs[6])  ||
                     !mxIsDouble(prhs[6])))
        mexErrMsgTxt("Input SEED must be a real scalar!");

    /****** Get row and column sizes for input image and weight matrix. ********/
    image_row=mxGetM(prhs[0]);
    image_col=mxGetN(prhs[0]);
    weight_row=mxGetM(prhs[1]);
    weight_col=mxGetN(prhs[1]);

    /****** Extract pointers to input data. ************************************/
    input_image=mxGetPr(prhs[0]);
    weight_matrix=mxGetPr(prhs[1]);
    if (nrhs>=3){
        hysterisis=mxGetPr(prhs[2]);
        hyst_row=mxGetM(prhs[2]);
        hyst_col=mxGetN(prhs[2]);
        }
    
    if (nrhs==2 || hyst_row==0 || hyst_col==0) {
        hysterisis=(double*)mxCalloc(1, sizeof(double));
        hysterisis[0]=Hysterisis_Constant;
        hyst_row=1;
        hyst_col=1;
        }
    
    if (nrhs>=4) direction=(int)mxGetScalar(prhs[3]);
    else         direction=-1;
    
    if (nrhs>=5) pair_combinations=mxGetPr(prhs[4]);
    else         pair_combinations=(double*)mxCalloc(weight_row*weight_col, sizeof(double));
    
    if (nrhs>=6) tolerances=mxGetPr(prhs[5]);
    else         tolerances=(double*)mxCalloc(weight_row*weight_col, sizeof(double));
    
    if (nrhs==7) srand((unsigned)mxGetScalar(prhs[6]));
    
    initialize_filter_pairs(pair_combinations, tolerances);

    /****** Create output variables and extract pointers to data. **************/
    output_dims[0]=image_row; output_dims[1]=image_col;
    plhs[0]=mxCreateLogicalArray(2, output_dims);
    output_data=mxGetLogicals(plhs[0]);
    if (nlhs==2){
        plhs[1]=mxCreateDoubleMatrix(image_row, image_col, mxREAL);
        error_image=mxGetPr(plhs[1]);
        }
    else error_image=(double*)mxCalloc(image_row*image_col, sizeof(double));
    output_image=mxCalloc(image_row*image_col, sizeof(double));
    
    /****** Call Error Diffusion function. *************************************/
    error_diffusion_with_hysterisis(output_image, error_image, input_image);
    for (n=0; n<image_col; n++){
        for (m=0; m<image_row; m++){
            output_data[m+n*image_row]=output_image[m+n*image_row]>=0.5;
            }
        }
    return;
}
