/****************************************************************************************/
/*                                                                                      */
/* HaLFtone MaSK                                                                        */
/*      This function converts a continuous-tone original image to binary by use        */
/*		of a pixelwise comparison between the original and a threshold array or mask.   */
/*	    Pixels of the original with intensities greater than their corresponding pixel  */
/*      in the mask are turned on (set to 1) while smaller intensities are turned off   */
/*		(set to 0). For large images, the dither array is tiled end-to-end.             */
/*                                                                                      */
/* Synopsis:                                                                            */
/*      Y=hlfmsk(X, M)                                                                  */
/*              X=Continuous tone original of type double or uint8.                     */
/*              M=Threshold array of same type as X.                                    */
/*                X and M can be composed of multiple channels, and if X has more       */
/*                channels than M, M is tiled along the C (color) axis.                 */
/*				Y=Resulting halftone image of type uint8 (logical).                     */
/*                                                                                      */
/* Dr. Daniel Leo Lau                                                                   */
/* Copyright June 13, 2000                                                              */
/*                                                                                      */
/****************************************************************************************/ 

#include "mex.h"

int imageRow=0;     // stores the size, in rows, of the input image
int imageCol=0;     // stores the size, in cols, of the input image
int imageChn=0;     // stores the size, in chns, of the input image

int maskRow=0;
int maskCol=0;
int maskChn=0;

/*************************************************************************/ 
/* mexFUNCTION                                                           */
/* Gateway routine for use with MATLAB.                                  */
/*************************************************************************/ 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    unsigned char *inputImage, *maskArray, *outputImage;
    int number_of_dims, output_dims[3], m, mp, n, np, r, rp;
    const int *dim_array;
      
    if (nrhs!=2)
        mexErrMsgTxt("DITHER requires two input arguments!");
    else if (!mxIsNumeric(prhs[0]) ||
              mxIsComplex(prhs[0]) ||
              mxIsSparse(prhs[0])  ||
             !mxIsUint8(prhs[0]))
        mexErrMsgTxt("Input X must be a real matrix of type uint8!");
    else if (!mxIsNumeric(prhs[1]) ||
              mxIsComplex(prhs[1]) ||
              mxIsSparse(prhs[1])  ||
             !mxIsUint8(prhs[1]))
        mexErrMsgTxt("Input Msk must be a real matrix of type uint8!");

    /****** Get row and column sizes for input image *********************/
	number_of_dims=mxGetNumberOfDimensions(prhs[0]);
	if (number_of_dims==2){
        imageRow=mxGetM(prhs[0]);
        imageCol=mxGetN(prhs[0]);
        imageChn=1;
    } else if (number_of_dims==3){
        dim_array=mxGetDimensions(prhs[0]);
        imageRow=dim_array[0];
        imageCol=dim_array[1];
        imageChn=dim_array[2];
    } else {
        mexErrMsgTxt("Input image X must 2 or 3 dimensional!");
    }
    inputImage=(unsigned char*)mxGetPr(prhs[0]);

    /****** Get row and column sizes for threshold array *********************/
	number_of_dims=mxGetNumberOfDimensions(prhs[1]);
	if (number_of_dims==2){
        maskRow=mxGetM(prhs[1]);
        maskCol=mxGetN(prhs[1]);
        maskChn=1;
    } else if (number_of_dims==3){
        dim_array=mxGetDimensions(prhs[1]);
        maskRow=dim_array[0];
        maskCol=dim_array[1];
        maskChn=dim_array[2];
    } else {
        mexErrMsgTxt("Threshold array must 2 or 3 dimensional!");
    }
    maskArray=(unsigned char*)mxGetPr(prhs[1]);

    /****** Create output variables and extract pointers to data. **************/
	output_dims[0]=imageRow;
    output_dims[1]=imageCol;
    output_dims[2]=imageChn;
    plhs[0]=mxCreateLogicalArray(3, output_dims);
    outputImage=(unsigned char*)mxGetPr(plhs[0]);
    
    for (r=0; r<imageChn; r++){
        rp=r%maskChn;
        for (n=0; n<imageCol; n++){
            np=n%maskCol;
            for (m=0; m<imageRow; m++){
                mp=m%maskRow;
                outputImage[m+n*imageRow+r*imageRow*imageCol]=
                            (inputImage[m+n*imageRow+r*imageRow*imageCol]>
                              maskArray[mp+np*maskRow+rp*maskRow*maskCol]);
            }
        }
    }
    
    return;
}
