/****************************************************************************************/
/*
/* LAU
/*	This function performs the error diffusion algorithm with output dependent feedback
/*  using 2 error and 2 balanced hysteresis filter weights.  This function allows the user
/*  to define the hysteresis parameter at each pixel of the input image as either a
/*  constant or a same sized matrix such that the hysteresis parameter used at pixel "n"
/*	is the value "h[n]".
/*
/* Synopsis:
/*	Y=lau(X, H)
/*		Y=Output image (binary)
/*		X=Input image, monochrome of type double
/*		H=Hysteresis matrix such that H[n] is the hysteresis parameter used to quantize X[n].
/*
/*
/* Daniel Leo Lau
/* Copyright February 22, 1997
/*
/****************************************************************************************/ 
#include <math.h>
#include "mex.h"
#define White_Value 		1.0
#define Threshold   		0.0
#define Hysterisis_Constant  	1.0

int direction=-1;
int image_row;
int image_col;
int hyst_row;
int hyst_col;
double *hysterisis;

/*******************************************************************************/ 
/*******************************************************************************/ 
/*******************************************************************************/ 
double lookup(double x_pt)
{
	double x_coor[30]={0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 
	                   1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9,
	                   2.0, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9};
	double y_coor[30]={0.430, 0.430, 0.435, 0.441, 0.454, 0.530, 0.535, 0.536, 0.540,
	                   0.544, 0.556, 0.569, 0.580, 0.595, 0.604, 0.616, 0.622, 0.629,
	                   0.637, 0.642, 0.648, 0.651, 0.658, 0.660, 0.669, 0.674, 0.678,
	                   0.683, 0.688, 0.692};
	int length=30;
	int m;
	double out;

	if (x_pt<x_coor[0]) return(y_coor[0]);
	for (m=1; m<length; m++){
	    if (x_coor[m-1]<=x_pt && x_pt<x_coor[m]){
	    	out=(x_pt-x_coor[m-1])*(y_coor[m]-y_coor[m-1]);
	    	out=out/(x_coor[m]-x_coor[m-1])+y_coor[m-1];
	    	return(out);
		}
	}
	return(y_coor[length-1]);
}

/*******************************************************************************/ 
/*******************************************************************************/ 
/*******************************************************************************/ 
void error_diffusion_with_hysterisis(double *output_image, double *error_image, double *input_image)
{	int m_p, n_p, r, t, d;
	int m, n;
        double A1, A2;

	for (m=0; m<image_row; m++){
	  d=(int)pow((float)direction, (float)m);
	  if (d==1) n=0; else n=image_col-1;
	  while(n>=0 && n<image_col){
            A1=lookup(hysterisis[(m%hyst_row)+(n%hyst_col)*hyst_row]);
            A2=1.0-A1;

	    output_image[m+n*image_row]=0.0;
            if (m>0)
              output_image[m+n*image_row]+=A1*output_image[(m-1)+n*image_row];
            if ((n-d)>=0 && (n-d)<image_col) 
              output_image[m+n*image_row]+=A2*output_image[m+(n-d)*image_row];

	    output_image[m+n*image_row]*=hysterisis[(m%hyst_row)+(n%hyst_col)*hyst_row];
	    output_image[m+n*image_row]+=input_image[m+n*image_row];
            output_image[m+n*image_row]+=error_image[m+n*image_row];

	    if (output_image[m+n*image_row] >= Threshold)
	      output_image[m+n*image_row]=1.0;
	    else
	      output_image[m+n*image_row]=0.0;

	    error_image[m+n*image_row]+=input_image[m+n*image_row];
            error_image[m+n*image_row]-=output_image[m+n*image_row]*White_Value;

            if ((m+1)<image_row)
              error_image[(m+1)+n*image_row]+=error_image[m+n*image_row]*0.5;
            if ((n+d)>=0 && (n+d)<image_col)
              error_image[m+(n+d)*image_row]+=error_image[m+n*image_row]*0.5;
 
	    n+=d;
	    }
	  }
	return;
}

/*******************************************************************************/ 
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/ 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	unsigned char *output_data;
	double *output_image, *error_image, *input_image;
	int m, n, output_dims[2];

    /****** Check for errors in user's call to function. ***********************/
    if (nrhs!=2)
            mexErrMsgTxt("LAU requires two input arguments!");
    else if (nlhs!=1)
            mexErrMsgTxt("LAU returns one output argument!");
    else if (!mxIsNumeric(prhs[0]) ||
              mxIsComplex(prhs[0]) ||
              mxIsSparse(prhs[0])  ||
             !mxIsDouble(prhs[0]))
            mexErrMsgTxt("Input X must be a real matrix of type double!");
    else if (!mxIsNumeric(prhs[1]) ||
              mxIsComplex(prhs[1]) ||
              mxIsSparse(prhs[1])  ||
             !mxIsDouble(prhs[1]))
            mexErrMsgTxt("Input H must be a real matrix of type double!");

    /****** Get row and column sizes for input image and weight matrix. ********/
    image_row=mxGetM(prhs[0]);
    image_col=mxGetN(prhs[0]);
	input_image=mxGetPr(prhs[0]);

	hyst_row=mxGetM(prhs[1]);
	hyst_col=mxGetN(prhs[1]);
	hysterisis=mxGetPr(prhs[1]);
	if (hyst_row==0 || hyst_col==0) {
		hysterisis=(double*)mxCalloc(1, sizeof(double));
		hysterisis[0]=Hysterisis_Constant;
		hyst_row=1;
		hyst_col=1;
	}

    /****** Create output variables and extract pointers to data. **************/
	output_dims[0]=image_row; output_dims[1]=image_col;
    plhs[0]=mxCreateLogicalArray(2, output_dims);
	output_data=(unsigned char*)mxGetPr(plhs[0]);
	output_image=mxCalloc(image_row*image_col, sizeof(double));
	error_image=(double*)mxCalloc(image_row*image_col, sizeof(double));

	error_diffusion_with_hysterisis(output_image, error_image, input_image);
	for (n=0; n<image_col; n++){
		for (m=0; m<image_row; m++){
			output_data[m+n*image_row]=output_image[m+n*image_row]>=0.5;
		}
	}
    return;
}
