#include <stdlib.h>
#include <math.h>
#include "mex.h"
#define pi 3.14159265358979
#define TOL 1e-32
#define max_num_pts 65536
#define NUM_ITERATIONS 500
#define NUM_LOOPS 1000

double TEMPERATURE=0.00005;
int image_row;
int image_col;

double *HVS;
int HVS_size=0;
int HVS_radius=0;

typedef struct {  
        int row;
        int col;
        double cost;
} cost_array_node;

void initialize_HVS_filter(double);
void hpsort_cost_array(int, cost_array_node*);
void initialize_cost_image(double*, unsigned char*);
void update_images(unsigned char*, double*, cost_array_node, int);
int vac(unsigned char*, double);

/********************************************************************************/
/*                                                                              */
/********************************************************************************/
void initialize_HVS_filter(double sigma)
{
	int m, n;
	double r, a, S=0.0;
	
	while (exp(-1.0*HVS_radius/(2*sigma*sigma))>=0.01) HVS_radius++;
	if (HVS_radius>((double)image_row/2.0)) HVS_radius=image_row/2;
	HVS_size=2*HVS_radius+1;
        /*mexPrintf("HVS radius=%d\n", HVS_radius);*/
	
	HVS=(double*)mxCalloc(HVS_size*HVS_size, sizeof(double));
	for (m=0; m<HVS_size; m++){
		for (n=0; n<HVS_size; n++){
			r=(m-HVS_radius)*(m-HVS_radius)+(n-HVS_radius)*(n-HVS_radius);
			a=atan2(m,n);
			HVS[m+n*HVS_size]=exp(-1.0*r/(2*sigma*sigma));
			S+=HVS[m+n*HVS_size];
		}
	}
	for (m=0; m<HVS_size*HVS_size; m++) HVS[m]=HVS[m]/S;
	
	return;
}

/********************************************************************************/
/*                                                                              */
/********************************************************************************/
void hpsort_cost_array(int n, cost_array_node *ra)
{
	int i, ir, j, l;
	cost_array_node rra;

	if (n<2) return;
	l=(n >> 1)+1;
	ir=n;
	for(;;){
		if (l>1){ --l; rra=ra[l-1]; }
		else {
			rra=ra[ir-1]; ra[ir-1]=ra[0];
			if (--ir==1){ra[0]=rra; break;}
		}
		i=l; j=l+l;
		while(j<=ir){
			if (j < ir && ra[j-1].cost < ra[j].cost)j++;
			if (rra.cost < ra[j-1].cost){
				ra[i-1]=ra[j-1]; i=j; j<<=1;
			}
			else j=ir+1;
		}
		ra[i-1]=rra;
	}
	return;
}

/********************************************************************************/
/*                                                                              */
/********************************************************************************/
void initialize_cost_image(double cost_image[], unsigned char output_image[])
{
	int r, s, m, n, dm, dn;
	double total_error=0.0;
	
	for (m=0; m<image_row; m++){
		for (n=0; n<image_col; n++){
			if (output_image[m+n*image_row]==1){
				for (r=-HVS_radius; r<=HVS_radius; r++){
					for (s=-HVS_radius; s<=HVS_radius; s++){
						dm=m+r;
						dn=n+s;
						while (dm<0) dm+=image_row;
						while (dm>=image_row) dm-=image_row;
						while (dn<0) dn+=image_col;
						while (dn>=image_col) dn-=image_col;
						cost_image[dm+dn*image_row]+=HVS[(r+HVS_radius)+(s+HVS_radius)*HVS_size];
					}
				}
			}
		}
	}
	return;
}

/********************************************************************************/
/*                                                                              */
/********************************************************************************/
void update_images(unsigned char output_image[], double cost_image[],
                     cost_array_node single_node, int one_or_zero)
{
	int r, s, dm, dn, m;
	
	output_image[single_node.row+single_node.col*image_row]=(unsigned char)one_or_zero;
	if (one_or_zero==0){
		for (r=-HVS_radius; r<=HVS_radius; r++){
			for (s=-HVS_radius; s<=HVS_radius; s++){
				dm=single_node.row+r;
				dn=single_node.col+s;
				while (dm<0) dm+=image_row;
				while (dm>=image_row) dm-=image_row;
				while (dn<0) dn+=image_col;
				while (dn>=image_col) dn-=image_col;
				cost_image[dm+dn*image_row]-=HVS[(r+HVS_radius)+(s+HVS_radius)*HVS_size];
			}
		}
	}
	else {
		for (r=-HVS_radius; r<=HVS_radius; r++){
			for (s=-HVS_radius; s<=HVS_radius; s++){
				dm=single_node.row+r;
				dn=single_node.col+s;
				while (dm<0) dm+=image_row;
				while (dm>=image_row) dm-=image_row;
				while (dn<0) dn+=image_col;
				while (dn>=image_col) dn-=image_col;
				cost_image[dm+dn*image_row]+=HVS[(r+HVS_radius)+(s+HVS_radius)*HVS_size];
			}
		}
	}
	return;
}

/********************************************************************************/
/*                                                                              */
/********************************************************************************/
int vac(unsigned char output_image[], double intensity)
{
	cost_array_node *one_list, *zero_list, single_node;
	double *cost_image;
	int m, n, r, s, num_ones, num_zeros, high_one, low_zero, num_iterations;
	
	num_ones=(int)(image_row*image_col*intensity);
	num_zeros=(image_row*image_col)-num_ones;
	
	/*** INITIALIZE OUTPUT_IMAGE AS A WHITE NOISE PROCESS ***********************/
	one_list=(cost_array_node*)mxCalloc(image_row*image_col, sizeof(cost_array_node));
	for (m=0; m<image_row; m++){
		for (n=0; n<image_col; n++){
			one_list[m+n*image_row].row=m;
			one_list[m+n*image_row].col=n;
            one_list[m+n*image_row].cost=(double)rand()/(double)RAND_MAX;
		}
	}
	hpsort_cost_array(image_row*image_col, one_list);
	for (m=0; m<num_ones; m++){
		output_image[one_list[m].row+one_list[m].col*image_row]=1;
	}
	mxFree(one_list);
	
	/*** INITIALIZE COST IMAGE **************************************************/
	cost_image=(double*)mxCalloc(image_row*image_col, sizeof(double));
	initialize_cost_image(cost_image, output_image);

	/*** INITIALIZE ONE AND ZERO LISTS ******************************************/
	one_list=(cost_array_node*)mxCalloc(num_ones, sizeof(cost_array_node));
	zero_list=(cost_array_node*)mxCalloc(num_zeros, sizeof(cost_array_node));
	r=0; s=0;
	for (m=0; m<image_row; m++){
		for (n=0; n<image_col; n++){
			if (output_image[m+n*image_row]==1){
				one_list[r].row=m;
				one_list[r].col=n;
				r++;
			}
			else if (output_image[m+n*image_row]==0){
				zero_list[s].row=m;
				zero_list[s].col=n;
				s++;
			}			
		}
	}
	num_ones=r;
	num_zeros=s;
	/*mexPrintf("Num Ones=%d, Num Zeros=%d\n\n", r, s);*/
	
	/**************************************************************************/
	/**** VOID AND CLUSTER STEPS **********************************************/
	/**************************************************************************/
	while(1){
		/*** REMOVE ONE WITH HIGHEST CONCENTRATION ****************************/
		high_one=0;
		for (m=1; m<num_ones; m++){
			if (cost_image[one_list[m].row+one_list[m].col*image_row] >
			    cost_image[one_list[high_one].row+one_list[high_one].col*image_row])
				high_one=m;
		}
		single_node=one_list[high_one];

		/*** UPDATE CONCENTRATION (COST) MATRIX *******************************/
		update_images(output_image, cost_image, single_node, 0);

		/*** REMOVE ZERO WITH LOWEST CONCENTRATION ****************************/
		low_zero=0;
		for (m=1; m<num_zeros; m++){
			if (cost_image[zero_list[m].row+zero_list[m].col*image_row] <
			    cost_image[zero_list[low_zero].row+zero_list[low_zero].col*image_row])
				low_zero=m;
		}
		/*mexPrintf("Low Concentration=%f\n", cost_image[zero_list[low_zero].row+
		  zero_list[low_zero].col*image_row]);*/

		/*** DOES VAC CONVERGE ON CURRENT ITERATION ***************************/
		if (cost_image[single_node.row+single_node.col*image_row] <=
		    cost_image[zero_list[low_zero].row+zero_list[low_zero].col*image_row]){
			output_image[single_node.row+single_node.col*image_row]=(unsigned char)1;
			break;
		}
		update_images(output_image, cost_image, zero_list[low_zero], 1);
		one_list[high_one]=zero_list[low_zero];
		zero_list[low_zero]=single_node;
		num_iterations++;
	}
	
    /* FREE CALLOCED SPACE AND RETURN ******************************************/
	mxFree(cost_image);
	return(num_iterations);
}

/*******************************************************************************/ 
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/ 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	unsigned char *output_image;
	double *error_image, gray_level, intensity, sigma, *num_iterations_data;
	int output_dims[2], m;
        
	if (!((nrhs>=2 && nrhs<=3) && (nlhs>=1 && nlhs<=2)))
		mexErrMsgTxt("Improper number of input/output arguments!");

	if (mxGetM(prhs[0])!=1 || mxGetN(prhs[0])!=1 || mxIsComplex(prhs[0]))
		mexErrMsgTxt("Input N must be a real scalar!");
    
	image_row=(int)mxGetScalar(prhs[0]);
	image_col=(int)mxGetScalar(prhs[0]);
	
    if (image_row<16)
		mexErrMsgTxt("Input N must be atleast 16!");
	
	if (mxGetM(prhs[1])!=1 || mxGetN(prhs[1])!=1 || mxIsComplex(prhs[1]))
		mexErrMsgTxt("Input g must be a real scalar!");
	
    gray_level=mxGetScalar(prhs[1]);
	
    if (gray_level<=0.0 || gray_level>=1.0)
		mexErrMsgTxt("Input g must be greater than 0 and less than 1!");
	else if (gray_level<=0.5) intensity=gray_level;
	else intensity=1.0-gray_level;
	
	if (nrhs>2){
		if (mxGetM(prhs[2])!=1 || mxGetN(prhs[2])!=1 || mxIsComplex(prhs[2]))
			mexErrMsgTxt("Input sigma must be a real scalar!");
		else
			sigma=mxGetScalar(prhs[2]);
	}
	else sigma=1.50;
	initialize_HVS_filter(sigma);
		
    output_dims[0]=image_row; output_dims[1]=image_col;
    plhs[0]=mxCreateLogicalArray(2, output_dims);
    output_image=(unsigned char*)mxGetPr(plhs[0]);

	if (nlhs==2){
		plhs[1]=mxCreateDoubleMatrix(1, 1, mxREAL);
		num_iterations_data=(double*)mxGetPr(plhs[1]);
		num_iterations_data[0]=(double)vac(output_image, intensity);
	}
	else vac(output_image, intensity);
	if (gray_level < 1.0 && gray_level > 0.5){
		for (m=0; m<image_row*image_col; m++){
			output_image[m]=!output_image[m];
		}
	}

    mxFree(HVS);
	return;
}

