/****************************************************************************************/
/*                                                                                      */
/* Dot Gain Model                                                                       */
/*      This function models the output of a binary display device using either the     */
/*      round circular dot model of Pappas, the stochastic dot model of Allebach or     */
/*      the stochastic dot model of Lau.                                                */
/*                                                                                      */
/* Synopsis:                                                                            */
/*      [H, P] = dgm(X, model, output, arg1, arg2, arg3, arg4)                          */
/*                X => The input halftone image to be printed                           */
/*            model => 'pappas' for the round circular dot model of Pappas              */
/*                         arg1 => the ratio of the diameter of the printer dot to the  */
/*                                 minimum distance between output pixels               */
/*                         arg2 => the sub-sampling rate                                */
/*                     'allebach' for the stochastic model of Allebach                  */
/*                         arg1 => the beam width                                       */
/*                         arg2 => the sub-sampling rate                                */
/*                     'lau' for the stochastic model of Lau                            */
/*                         arg1 => the beam width                                       */
/*                         arg2 => the sub-sampling rate                                */
/*                         arg3 => X-cooridinates of transfer function                  */
/*                         arg4 => Y-cooridinates of transfer function                  */
/*           output => Output type (default="pixels")                                   */
/*                     'pixels' for output of the individual toner particles            */
/*                     'average' for output of the average intensity of each pixel      */
/*                                                                                      */
/* Daniel Leo Lau                                                                       */
/* Copyright November 15, 1998                                                          */
/*                                                                                      */
/****************************************************************************************/ 
#include <stdlib.h>
#include <math.h>
#include "mex.h"
#define TOL  0.001
#define PIXELS   0
#define AVERAGE  1
#define PAPPAS   0
#define ALLEBACH 1
#define LAU      2

int image_row;
int image_col;
int D;
int MODEL;
int FLAG;
double alpha;
double *x_coor;
double *y_coor;
int length;

/*******************************************************************************/ 
/*******************************************************************************/ 
/*******************************************************************************/ 
double lookup(double x_pt)
{
	int m;

	for (m=1; m<length; m++){
		if (x_coor[m-1]<=x_pt && x_pt<x_coor[m]){
            return((x_pt-x_coor[m-1])*(y_coor[m]-y_coor[m-1])/
                   (x_coor[m]-x_coor[m-1])+y_coor[m-1]);
		}
	}
	return(y_coor[length-1]);
}

/*******************************************************************************/ 
/*******************************************************************************/ 
/*******************************************************************************/ 
void sdm(unsigned char *output_image, double *prob_image, unsigned char *input_image)
{
        int m, n, r, s, s_low, s_high, r_low, r_high, radius;
        double z, res;

        if (MODEL==PAPPAS){
            radius=(int)(alpha*D/2.0);
	}
        else{
            radius=0;
            while(exp(-alpha*radius*radius/(D*D))>TOL) radius++;
	}

        if (MODEL==PAPPAS){
          for (n=0;n<image_col;n++){
	    for (m=0;m<image_row;m++){
              if (input_image[m+n*image_row]==(unsigned char)0){
                r_low=(m+0.5)*D-radius;  if (r_low<0) r_low=0;
                r_high=(m+0.5)*D+radius; if (r_high>image_row*D) r_high=image_row*D;
                s_low=(n+0.5)*D-radius;  if (s_low<0) s_low=0;
                s_high=(n+0.5)*D+radius; if (s_high>image_col*D) s_high=image_col*D;
                for (s=s_low;s<s_high;s++){
                  for (r=r_low;r<r_high;r++){
                    z=(double)(r-((double)m+0.5)*D)*(double)(r-((double)m+0.5)*D)+
                      (double)(s-((double)n+0.5)*D)*(double)(s-((double)n+0.5)*D);
                    if (z<=(radius*radius))
                         prob_image[r+s*image_row*D]=1.0;
		  }
	        }
	      }
	    }
	  }
	}
        else if (MODEL==ALLEBACH){
          for (m=0; m<image_row*image_col*D*D; m++) prob_image[m]=1.0;
          for (n=0;n<image_col;n++){
	    for (m=0;m<image_row;m++){
              if (input_image[m+n*image_row]==(unsigned char)0){
                r_low=(m+0.5)*D-radius;  if (r_low<0) r_low=0;
                r_high=(m+0.5)*D+radius; if (r_high>image_row*D) r_high=image_row*D;
                s_low=(n+0.5)*D-radius;  if (s_low<0) s_low=0;
                s_high=(n+0.5)*D+radius; if (s_high>image_col*D) s_high=image_col*D;
                for (s=s_low;s<s_high;s++){
                  for (r=r_low;r<r_high;r++){
                    z=(double)(r-((double)m+0.5)*D)*(double)(r-((double)m+0.5)*D)+
                      (double)(s-((double)n+0.5)*D)*(double)(s-((double)n+0.5)*D);
                    prob_image[r+s*image_row*D]*=(1.0-exp(-alpha*z/(D*D)));
                  }
                }
	      }
    	    }
	  }
          for (m=0; m<image_row*image_col*D*D; m++) prob_image[m]=1.0-prob_image[m];
	}
        else if (MODEL==LAU){
          for (n=0;n<image_col;n++){
	    for (m=0;m<image_row;m++){
              if (input_image[m+n*image_row]==(unsigned char)0){
                r_low=(m+0.5)*D-radius;  if (r_low<0) r_low=0;
                r_high=(m+0.5)*D+radius; if (r_high>image_row*D) r_high=image_row*D;
                s_low=(n+0.5)*D-radius;  if (s_low<0) s_low=0;
                s_high=(n+0.5)*D+radius; if (s_high>image_col*D) s_high=image_col*D;
                for (s=s_low;s<s_high;s++){
                  for (r=r_low;r<r_high;r++){
                    z=(double)(r-((double)m+0.5)*D)*(double)(r-((double)m+0.5)*D)+
                      (double)(s-((double)n+0.5)*D)*(double)(s-((double)n+0.5)*D);
                    prob_image[r+s*image_row*D]+=exp(-alpha*z/(D*D));
                  }
		}
    	      }
	    }
	  }
          for (s=0;s<image_col*D;s++){
	    for (r=0;r<image_row*D;r++){
              prob_image[r+s*image_row*D]=lookup(prob_image[r+s*image_row*D]);
            }
          }
	}

        if (FLAG==PIXELS){
          for (s=0;s<image_col*D;s++){
            for (r=0;r<image_row*D;r++){
              output_image[r+s*image_row*D]=((double)rand()/(double)RAND_MAX) >
                                            prob_image[r+s*image_row*D];
	      }
	    }
          }
        else{
          for (s=0;s<image_col;s++){
	    for (r=0;r<image_row;r++){
              res=0;
              for (m=0; m<D; m++){
	        for (n=0; n<D; n++){
                  res+=prob_image[(r*D+m)+(s*D+n)*image_row*D];
                }
	      }
              output_image[r+s*image_row]=(unsigned char)(255.0*(1.0-(res/(double)(D*D))));
	    }
	  }
	}
        return;
}

/*******************************************************************************/ 
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/ 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	unsigned char *input_image, *output_image;
	double *prob_image;
	int m, n, output_dims[2];
	mxChar *output_str, *model_str;

        if (nrhs<2)
                mexErrMsgTxt("Too few input arguments in DGM!");
	else if (nlhs!=1 && nlhs!=2)
                mexErrMsgTxt("DGM returns one or two output arguments!");
        else if (!mxIsNumeric(prhs[0]) ||
                  mxIsComplex(prhs[0]) ||
                  mxIsSparse(prhs[0])  ||
                 !mxIsUint8(prhs[0])   ||
                  mxGetM(prhs[0])<1    ||
                  mxGetN(prhs[0])<1) 
                mexErrMsgTxt("Input X must be a real matrix of type uint8 (logical)!");
        else if (!mxIsChar(prhs[1]) ||
                 (mxGetM(prhs[1])*mxGetN(prhs[1]))<1)
                mexErrMsgTxt("MODEL must be character string!");
        else if (!mxIsChar(prhs[2]))
                mexErrMsgTxt("OUTPUT must be character string!");

        model_str=(mxChar*)mxGetPr(prhs[1]);
        if (model_str[0]==112 || model_str[0]==80){     /* PAPPAS MODEL */
                MODEL=PAPPAS;
                if (nrhs!=4 && nrhs!=5)
                     mexErrMsgTxt("Pappas model requires 4 or 5 input arguments!");
                else if (nlhs!=1)
                     mexErrMsgTxt("Pappas model returns exactly 1 output argument!");
                else if (!mxIsNumeric(prhs[3]) ||
                          mxIsComplex(prhs[3]) ||
                          mxIsSparse(prhs[3])  ||
                        !(mxGetM(prhs[3])==1)  ||
                        !(mxGetN(prhs[3])==1))
                        mexErrMsgTxt("Radius must be a real scalar!");
                alpha=mxGetScalar(prhs[3]);
                if (nrhs==5){
                    if (!mxIsNumeric(prhs[4]) ||
                         mxIsComplex(prhs[4]) ||
                         mxIsSparse(prhs[4])  ||
		       !(mxGetM(prhs[4])==1)  ||
                       !(mxGetN(prhs[4])==1))
                         mexErrMsgTxt("Subsampling rate must be a real scalar!");
                    else
                         D=(int)mxGetScalar(prhs[4]);
                    }
                else
                    D=6;
	        }
        else if (model_str[0]==65 || model_str[0]==97){     /* ALLEBACH MODEL */
                MODEL=ALLEBACH;
                if (nrhs!=4 && nrhs!=5)
                     mexErrMsgTxt("Allebach model requires 4 or 5 input arguments!");
                else if (!mxIsNumeric(prhs[3]) ||
                          mxIsComplex(prhs[3]) ||
                          mxIsSparse(prhs[3])  ||
		        !(mxGetM(prhs[3])==1)  ||
		        !(mxGetN(prhs[3])==1))
                          mexErrMsgTxt("Alpha must be a real scalar!");
                alpha=mxGetScalar(prhs[3]);
                if (nrhs==5){
                    if (!mxIsNumeric(prhs[4]) ||
                         mxIsComplex(prhs[4]) ||
                         mxIsSparse(prhs[4])  ||
		       !(mxGetM(prhs[4])==1)  ||
                       !(mxGetN(prhs[4])==1))
                         mexErrMsgTxt("Subsampling rate must be a real scalar!");
                    else
                         D=(int)mxGetScalar(prhs[4]);
                    }
                else
                         D=6;
	        }
        else if (model_str[0]==76 || model_str[0]==108){     /* LAU MODEL */
                MODEL=LAU;
                if (nrhs!=7)
                     mexErrMsgTxt("Lau model requires 7 input arguments!");
                else if (!mxIsNumeric(prhs[3]) ||
                          mxIsComplex(prhs[3]) ||
                          mxIsSparse(prhs[3])  ||
                        !(mxGetM(prhs[3])==1)  ||
                        !(mxGetN(prhs[3])==1))
                        mexErrMsgTxt("Alpha must be a real scalar!");
                else if (!mxIsNumeric(prhs[4]) ||
                          mxIsComplex(prhs[4]) ||
                          mxIsSparse(prhs[4])  ||
		        !(mxGetM(prhs[4])==1)  ||
                        !(mxGetN(prhs[4])==1))
                        mexErrMsgTxt("Subsampling rate must be a real scalar!");
                else if (!mxIsNumeric(prhs[5]) ||
                          mxIsComplex(prhs[5]) ||
                          mxIsSparse(prhs[5])  ||
	        	!(mxGetM(prhs[5])*mxGetN(prhs[5])>1))
                        mexErrMsgTxt("X cooridinates must be a real vector!");
                else if (!mxIsNumeric(prhs[6]) ||
                          mxIsComplex(prhs[6]) ||
                          mxIsSparse(prhs[6])  ||
	        	!(mxGetM(prhs[6])*mxGetN(prhs[6])>1))
                        mexErrMsgTxt("Y cooridinates must be a real vector!");
                alpha=mxGetScalar(prhs[3]);
                D=(int)mxGetScalar(prhs[4]);
                length=mxGetM(prhs[5])*mxGetN(prhs[5]);
                if (mxGetM(prhs[6])*mxGetN(prhs[6]) != length){
                        mexErrMsgTxt("Y and X cooridinates must be same length!");
	                }
                x_coor=mxGetPr(prhs[5]);
                for (m=1; m<length; m++){
	                if (x_coor[m]<x_coor[m-1])
                              mexErrMsgTxt("X cooridinates must be in ascending order!");
	                }
                y_coor=mxGetPr(prhs[6]);
                for (m=0; m<length; m++){
	                if (y_coor[m]<0 || y_coor[m]>1)
                               mexErrMsgTxt("Y cooridinates must be within range [0,1]!");
	                }
	        }
        else
                mexErrMsgTxt("Invalid printer model in DGM!");
        image_row=mxGetM(prhs[0]);
        image_col=mxGetN(prhs[0]);
        input_image=(unsigned char*)mxGetPr(prhs[0]);
        output_str=(mxChar*)mxGetPr(prhs[2]);

        if (mxGetM(prhs[2])*mxGetN(prhs[2])<1){
             output_dims[0]=image_row;
             output_dims[1]=image_col;
             FLAG=AVERAGE;
             }
        else if (output_str[0]==80  || output_str[0]==112){
             output_dims[0]=image_row*D;
             output_dims[1]=image_col*D;
             FLAG=PIXELS;
             }
        else{output_dims[0]=image_row;
             output_dims[1]=image_col;
             FLAG=AVERAGE;
             }
             
        if (FLAG==PIXELS) {
            plhs[0]=mxCreateLogicalArray(2, output_dims);
            output_image=mxGetLogicals(plhs[0]);
        }
        else {
            plhs[0]=mxCreateNumericArray(2, output_dims, mxUINT8_CLASS, mxREAL);
            output_image=(unsigned char*)mxGetPr(plhs[0]);
        }
        
        if (nlhs==2){
          plhs[1]=mxCreateDoubleMatrix(image_row*D, image_col*D, mxREAL);
          prob_image=mxGetPr(plhs[1]);
	  }
        else {
          prob_image=(double*)mxCalloc(image_row*D*image_col*D, sizeof(double));
	  }
        sdm(output_image, prob_image, input_image);
        return;
}
